## Procedures for implementing the detailed decomposition procedure in
## Rothe (2012): "Decomposing the composition effect"

# Written by Christoph Rothe
# This version 01/25/2013


# Load required packages
library(foreign)
library(Hmisc)
library(mvtnorm)
library(copula)
library(survey)
library(xtable)
library(clusterGeneration)
library(corpcor)

# Define the Gini coeficient
 Gini <- function (x, w) {
    n <- length(x)
    w <- w/sum(w)
    G <- sum(x[order(x)] * 1:n * w[order(x)])
    G <- 2 * G/(n*sum(x[order(x)]  * w[order(x)]))
    G - 1 - (1/n)
 }
 
# Modify the function wtd.quantile such that it has the properties described
# in the paper.

wtd.quantile <- function (x, weights = NULL, probs = c(0, 0.25, 0.5, 0.75, 1),
    type = c("quantile", "(i-1)/(n-1)", "i/(n+1)", "i/n"), normwt = FALSE,
    na.rm = TRUE)
{
    if (!length(weights))
        return(quantile(x, probs = probs, na.rm = na.rm))
    type <- match.arg(type)
    if (any(probs < 0 | probs > 1))
        stop("Probabilities must be between 0 and 1 inclusive")
    nams <- paste(format(round(probs * 100, if (length(probs) >
        1) 2 - log10(diff(range(probs))) else 2)), "%", sep = "")
    if (type == "quantile") {
        w <- wtd.table(x, weights, na.rm = na.rm, normwt = normwt,
            type = "list")
        x <- w$x
        wts <- w$sum.of.weights
        n <- sum(wts)
        order <- 1 + (n - 1) * probs
        low <- pmax(floor(order), 1)
        high <- pmin(low + 1, n)
        order <- order%%1
        allq <- approx(cumsum(wts), x, xout = c(low, high), method = "constant",
            f = 1, rule = 2)$y
        k <- length(probs)
        quantiles <- (1 - order) * allq[1:k] + order * allq[-(1:k)]
        names(quantiles) <- nams
        return(quantiles)
    }
    w <- wtd.Ecdf(x, weights, na.rm = na.rm, type = type, normwt = normwt)
    structure(approx(w$ecdf, w$x, xout = probs, rule = 2, f=1,method = "constant",ties="ordered")$y,
        names = nams)
}



# decomp.object is a function that estimates the copula and the conditional CDF.
# It is later used to create an object on which the actual decomposition is then
# performed.

decomp.object <- function(formula, covariate, dat, weights = NA) {

# Define the Gaussian (normal) copula
 norm.copula <- function(x, corrs) {
   pmvnorm(lower = rep(-Inf,length(x)), upper = qnorm(x),corr=corrs)[1]
 }


# Define an objective function to estimate the parameters of the Gaussian copula
 objective <- function(corrs,i,www) {
    cormat <- diag(dim(m.cdf[[i]])[2])
    cormat[upper.tri(cormat)] <- corrs
    cormat <- cormat + t(cormat) - diag(dim(m.cdf[[i]])[2])
    length(www)*wtd.mean((j.cdf[[i]] - apply(m.cdf[[i]],1,norm.copula, corrs=cov2cor(make.positive.definite(cormat))))^2, w=www, normwt=TRUE) + 100*sum(pmax(rep(0,length(corrs)),abs(corrs)-1))
 }

# get variable name
 outname <- all.vars(formula)[1]
 covname <- all.vars(formula)[-1]
 pure.covname <- all.vars(covariate)

# Define List objects
 m.cdf <- list(NA)
 j.cdf <- list(NA)
 n.corr <- list(NA)
 copula.para <- list(NA)
 norm.cop <- list(NA)
 cond.cdf <- list(NA)
 dat.modmat <- list(NA)
 dat.pure <- list(NA)
 y <- list(NA)
 n <- list(NA)
 new.dat <- list(NA)

 ygrid <- quantile(c(dat[[1]][,outname],dat[[2]][,outname]), c(seq(0.001,.999,length=101))  )

 dat.pure[[1]] <- model.matrix(covariate, data = dat[[1]])[,-1]
 dat.pure[[2]] <- model.matrix(covariate, data = dat[[2]])[,-1]

 pb <- txtProgressBar(min = 0, max = 6, style = 3)

 for (i in 1:2) {

  # determine sample size
  n[[i]] <- dim(dat[[i]])[1]

  # check for presence of weights. if false assign 1/n
  if (sum(is.na(weights[[i]]))>0) {weights[[i]] <- rep(1,n[[i]])}

  # Prepare some data for output later
  y[[i]] <- dat[[i]][,outname]

  ## Estimate Conditional CDFs!

   dat.modmat[[i]] <- model.matrix(formula, data = dat[[i]])[,-1]
   cond.cdf[[i]] <- matrix(NA,ncol=dim(dat.modmat[[i]])[2]+1,nrow=length(ygrid))

   for (j in 2:(length(ygrid)-1)){
     dat[[i]][,outname] <-  I(y[[i]] <= ygrid[j])
     design <- svydesign(~0, data=dat[[i]], weights=~weights[[i]])
     cond.cdf[[i]][j,] <- svyglm(formula, data=dat[[i]], design=design,
                                family=quasibinomial(link="probit"),
                                start=(if (exists("mod")) {coef(mod)} ),
                                control=list(epsilon=1e-8,maxit=50,trace=FALSE))$coef
     setTxtProgressBar(pb, 3*(i-1) + j/length(ygrid))
    }
   dat[[i]][,outname] <- y[[i]]


  ## Estimate Marginal CDFs
  ## Randomly determine at which points to evaluate the joint CDF (for calculation
  ## of MD estimator). Here, the evaluation is performed at 500 randomly selected
  ## points. With smaller datasets than the one used in the paper, one could choose
  ## a higher number here.
  ss <- sample(1:length(dat.pure[[i]][,1]), min(c(length(dat.pure[[i]][,1]),500)),prob=weights[[i]], replace=FALSE)

  m.cdf[[i]] <- dat.pure[[i]]
  for (j in 1:dim(dat.pure[[i]])[2]) {
    wecdf <- wtd.Ecdf(dat.pure[[i]][,j], w=weights[[i]], normwt=TRUE)
    m.cdf[[i]][,j] <- approx(wecdf$x[-1], wecdf$ecdf[-1], xout = dat.pure[[i]][,j],  f=1, yleft=0, yright=1, method = "constant")$y
  }
  m.cdf[[i]] <-  m.cdf[[i]][ss,]

 ## Estimate Copula Functions

 j.cdf[[i]] <- 0
 for (j in 1:length(ss)) {
   j.cdf[[i]][j] <- wtd.mean(apply(t(dat.pure[[i]])- dat.pure[[i]][ss[j],] <= 0, 2,prod), w=weights[[i]], normwt=TRUE)
    setTxtProgressBar(pb, 3*(i-1) + 1 + j/length(ss))
 }

 n.corr[[i]] <- dim(m.cdf[[i]])[2]*(dim(m.cdf[[i]])[2]-1)/2
 copula.para[[i]] <- optim( rep(0,n.corr[[i]]) , objective, i=i, www=weights[[i]][ss], control=list(trace=0,maxit=500))$par


 cormat <- diag(dim(m.cdf[[i]])[2])
 cormat[upper.tri(cormat)] <- copula.para[[i]]
 cormat <-  cov2cor(make.positive.definite(cormat + t(cormat) - diag(dim(m.cdf[[i]])[2])))


 ## Generate hypothetical data for decomposition later.
   new.dat[[i]] <- list(NA)
   new.dat[[i]][[1]] <- pnorm(rmvnorm(500000,sigma=cormat))
   new.dat[[i]][[2]] <- new.dat[[i]][[1]]
  for (l in 1:dim(new.dat[[i]][[1]])[2]) {
      new.dat[[i]][[1]][,l] <- wtd.quantile(dat.pure[[1]][,l], new.dat[[i]][[1]][,l],type="i/n", w=weights[[1]], normwt=TRUE)
      new.dat[[i]][[2]][,l] <- wtd.quantile(dat.pure[[2]][,l], new.dat[[i]][[2]][,l],type="i/n", w=weights[[2]], normwt=TRUE)
      setTxtProgressBar(pb, 3*(i-1) + 2+l/dim(new.dat[[i]][[1]])[2])
    }
  }
 close(pb)

list(norm.cop, dat.pure, cond.cdf, ygrid, outname, formula, weights,  pure.covname, y, new.dat)

}


## the function decomp performs the computation of the various elements of the
## detailed decompositions. The arguments conditional, copula and marginals
## specify from which group the respective object is to be taken.

decomp <- function(decomp.obj, conditional = 1, copula =1, marginals = c(1,2)) {

 dat.pure <- decomp.obj[[2]]
 cond.cdf <- decomp.obj[[3]]
 ygrid <- decomp.obj[[4]]
 outname <- decomp.obj[[5]]
 formula <-  decomp.obj[[6]]
 weights <-  decomp.obj[[7]]
 y <- decomp.obj[[9]]
 new.dat <- decomp.obj[[10]]

 i <- conditional     ## conditional
 j <- copula          ## copula
 k <- marginals       ## marginal

  ## Create New Covariate Data (for numerical integration)

 newxdat <- new.dat[[j]][[1]]
  for (l in 1:dim(newxdat)[2]) {
    newxdat[,l] <- new.dat[[j]][[k[l]]][,l]
  }

 newxdat <- data.frame(rep(1,dim(newxdat)[1]),newxdat)
 colnames(newxdat) <- c(outname,colnames(dat.pure[[1]]))
 newevdat <- model.matrix(formula, data = newxdat)

 ## Compute Counterfactual Outcome CDF
 cdf.out <- 1/dim(dat.pure[[i]])[1]
 for (l in 2:(length(ygrid)-1)) {
  cdf.out[l] <- mean(pnorm(as.vector(newevdat%*%cond.cdf[[i]][l,])))
 }
 cdf.out[length(ygrid)] <- 1

 pseudo.data <- approx(x=sort(cdf.out), y=ygrid, xout=seq(0,1,length=50000), method="linear", rule=2)$y
 ## can compute summary measures from pseudo data.

c(mean(pseudo.data),
   quantile(pseudo.data,.9),
   quantile(pseudo.data,.5),
   quantile(pseudo.data,.1),
   quantile(pseudo.data,.9)-quantile(pseudo.data,.1),
   quantile(pseudo.data,.9)-quantile(pseudo.data,.5),
   quantile(pseudo.data,.5)-quantile(pseudo.data,.1),
   sd(pseudo.data),
   Gini(pseudo.data, w=rep(1,length(pseudo.data))))

}


## the function full.decomp computes the total group difference, structure and
## composition effect, dependence and marginal distributions effects, the direct
## contribution of differences in each marginal distribution, and all two way
## interaction terms. This is done for a variety of parameters: mean, several
## quantiles and quantile differences, the standard deviation, and the Gini
## coeficient.

full.decomp <- function(formula, covariate, dat, weights = NA) {


 test <<- decomp.object(formula, covariate, dat, weights)
 varnames  <- test[[8]]
 nvar <- length(varnames)

 v <- matrix(0,ncol=9,nrow=4+sum(nvar:1))

 v[1,] <- decomp(test,conditional = 2, copula =2, marginals = rep(2,nvar)) ## COULD ALSO USE RAW SAMPLE VALUE
 v[2,] <- decomp(test,conditional = 1, copula =2, marginals = rep(2,nvar)) ## COULD ALSO USE VALUE THAT CIRCUMVENTS COPULA ESTIMATION
 v[3,] <- decomp(test,conditional = 1, copula =1, marginals = rep(2,nvar))
 v[4,] <- decomp(test,conditional = 1, copula =1, marginals = rep(1,nvar)) ## COULD ALSO USE RAW SAMPLE VALUE

 base.marginals <-  rep(1,nvar)

 for (i in 1:nvar) {
  mm <- base.marginals
  mm[i] <- mm[i]+1
  v[4+i,] <- decomp(test,conditional = 1, copula =1, marginals = mm) -  v[4,]
 }

 for (i in 1:(nvar-1)) {
  for (j in 1:(nvar-i)) {
  mm <- base.marginals
  mm[c(i,i+j)] <- mm[c(i,i+j)]+c(1,1)
  v[(4+nvar)+ cumsum(c(0,((nvar-1)):2 ))[i]  + j, ] <- decomp(test,conditional = 1, copula =1, marginals = mm) - v[4,] - v[4+i,] - v[4+i+j,]
  varnames <- c(varnames,paste(varnames[c(i,i+j)],collapse =":"))
  }
 }

 out <- as.matrix(rbind(v[1,]-v[4,], v[1,]-v[2,], v[2,]-v[4,], v[2,]-v[3,], v[3,]-v[4,], v[-(1:4),] ))
 rownames(out) <- c("total","structure","composition", "dependence", "marginal",varnames)
 colnames(out)<- c("mean","q90","q50","q10","q9010","q9050","q5010","sd","gini")

 out

}
